/*
* Copyright (c) 2005 Erwin Coumans <www.erwincoumans.com>
*
* Permission to use, copy, modify, distribute and sell this software
* and its documentation for any purpose is hereby granted without fee,
* provided that the above copyright notice appear in all copies.
* Erwin Coumans makes no representations about the suitability 
* of this software for any purpose.  
* It is provided "as is" without express or implied warranty.
*/



#include "BulletCollision/CollisionDispatch/btCollisionWorld.h"

/*
Raytracer uses the Convex rayCast to visualize the Collision Shapes/Minkowski Sum.
Very basic raytracer, rendering into a texture.
*/

///Low level demo, doesn't include btBulletCollisionCommon.h

#include "LinearMath/btQuaternion.h"
#include "LinearMath/btTransform.h"
#include "GL_ShapeDrawer.h"
#include "GLDebugDrawer.h"

#include "Raytracer.h"
#include "GlutStuff.h"


#include "BulletCollision/NarrowPhaseCollision/btVoronoiSimplexSolver.h"
#include "BulletCollision/NarrowPhaseCollision/btSubSimplexConvexCast.h"
#include "BulletCollision/NarrowPhaseCollision/btGjkConvexCast.h"
#include "BulletCollision/NarrowPhaseCollision/btContinuousConvexCollision.h"

#include "BulletCollision/CollisionShapes/btSphereShape.h"
#include "BulletCollision/CollisionShapes/btMultiSphereShape.h"

#include "BulletCollision/CollisionShapes/btConvexHullShape.h"
#include "LinearMath/btAabbUtil2.h"
#include "BulletCollision/CollisionShapes/btBoxShape.h"
#include "BulletCollision/CollisionShapes/btCompoundShape.h"


#include "BulletCollision/CollisionShapes/btTetrahedronShape.h"
#include "BulletCollision/CollisionShapes/btConeShape.h"
#include "BulletCollision/CollisionShapes/btCylinderShape.h"
#include "BulletCollision/CollisionShapes/btMinkowskiSumShape.h"

#include "BulletCollision/CollisionDispatch/btDefaultCollisionConfiguration.h"
#include "BulletCollision/BroadphaseCollision/btAxisSweep3.h"
#include "BulletCollision/BroadphaseCollision/btDbvtBroadphase.h"

#include "BulletCollision/BroadphaseCollision/btBroadphaseProxy.h"
#include "BulletCollision/CollisionShapes/btScaledBvhTriangleMeshShape.h"
#include "BulletCollision/CollisionShapes/btUniformScalingShape.h"
#include "BulletCollision/CollisionShapes/btScaledBvhTriangleMeshShape.h"
//========================================================================================================================
// PUBLIC:
// This can be tweaked for speed vs precision (powers of two in theory, but others work as well...) 64 - 256 goes well on my system
static const int raytraceResolutionLow = 64;				// used when the camera is moving
static const int raytraceResolutionHigh = 256;				// used after the camera has stopped.
static const int framesInterval = 60;	// MUST BE >=2! It's the time (in frames) between the low-res and the hi-res switch, once keys are released. Bigger values prevent "false positives" and allow the user to be happier.
//========================================================================================================================


//========================================================================================================================
// PRIVATE:	(well, it should be better to embed them...)
static const int textureSize[2] = {raytraceResolutionLow,raytraceResolutionHigh};	// First resolution is used when the camera is moving, other when it's still. ( MUST BE first <= second ! )
static const int textureArea = raytraceResolutionHigh * raytraceResolutionHigh;
static unsigned char textureBuffer[raytraceResolutionHigh*raytraceResolutionHigh];	// We never resize it, so that the address of the elements is constant.	
GLuint glTextureId[2];	// Low - High
static const int textureMinFilter = GL_LINEAR; 
static const int textureMagFilter = GL_LINEAR;
//========================================================================================================================

// This is instead a very useful method IMO, but I don't know how fast it is.
SIMD_FORCE_INLINE static btVector3 GetAABBExtents(const btCollisionShape* shape)	{
		if (shape)	{
			btVector3 aabbMin,aabbMax;
			shape->getAabb(btTransform::getIdentity(),aabbMin,aabbMax);
			return (aabbMax - aabbMin);		
		}	
		return btVector3(0,0,0);
}



//================================================================
// class Raytracer
//================================================================

void	Raytracer::initPhysics()
{
	// Create texture used for raytrace:
	createTexture();
	// World creation
	initWorld();
	//=====================================================================================================
	// User space:
	btCollisionShape* myShape = new btCylinderShapeX(btVector3(6,4,4));
	m_collisionShapeGarbageCollector.push_back(myShape);	// Optional but recommended... it SHOULD handle child shapes as well... (it works behind the scenes for all bodies in the world too: that's why it's optional)
	btTransform T(btQuaternion::getIdentity(),m_cameraTargetPosition);	// Tip: place the center of your objects in m_cameraTargetPosition (should be zero)
	
	btCollisionObject* obj = new btCollisionObject();
	obj->setCollisionShape(myShape);
	obj->setWorldTransform(T);
	
	m_collisionWorld->addCollisionObject(obj);
	
	// Debug text :
	const btVector3 exts = GetAABBExtents(myShape);
	printf("Aabb FULL extents: (%1.2f,%1.2f,%1.2f)\n",exts.x(),exts.y(),exts.z());	
	//======================================================================================================

			
}

void	Raytracer::initWorld()	{
	m_collisionConfiguration = new btDefaultCollisionConfiguration();
	m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
	m_overlappingPairCache = new btDbvtBroadphase();
	m_collisionWorld = new btCollisionWorld(m_dispatcher,m_overlappingPairCache,m_collisionConfiguration);	
}
Raytracer::~Raytracer()	{

	//cleanup in the reverse order of creation/initialization

	//remove the rigidbodies from the dynamics world and delete them
	int i;
	for (i=m_collisionWorld->getNumCollisionObjects()-1; i>=0 ;i--)
	{
		btCollisionObject* obj = m_collisionWorld->getCollisionObjectArray()[i];
		if (!obj) continue;
		m_collisionShapeGarbageCollector.push_back(obj->getCollisionShape(),true);	// Tip for deleting shapes the user forgets to add...
		m_collisionWorld->removeCollisionObject( obj );
		delete obj;
	}
	m_collisionShapeGarbageCollector.clear();	// Useless...

	//delete collision world
	delete m_collisionWorld;

	//delete broadphase
	delete m_overlappingPairCache;

	//delete dispatcher
	delete m_dispatcher;

	delete m_collisionConfiguration;
	
	glDeleteTextures(2, &glTextureId[0]);
}


void Raytracer::createTexture(int textureIndex)	{
	if (!m_initialized)	{
		m_initialized = true;
		for (int i=0;i<textureArea;i++) textureBuffer[i]=255;	// Don't know if memset, memcpy, etc. are portable...
		glGenTextures(2, &glTextureId[0]);
		createTexture(0);
		createTexture(1);		
		return;
	}
	
	glBindTexture(GL_TEXTURE_2D,glTextureId[textureIndex]);
	
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, textureMagFilter);	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, textureMinFilter);	

	glTexImage2D(GL_TEXTURE_2D, 
		0, 
		GL_LUMINANCE, 
		textureSize[textureIndex],textureSize[textureIndex], 
		0, 
		GL_LUMINANCE, 
		GL_UNSIGNED_BYTE, 
		&textureBuffer[0]);
	
}
void Raytracer::updateTexture(int textureIndex)	{
	createTexture(textureIndex);	// Don't know if there are faster methods here... (glCopyTexSubImage2D() ?)
}

struct btSingleRayCallback : public btBroadphaseRayCallback	{

	btVector3	m_rayFromWorld;
	btVector3	m_rayToWorld;
	btTransform	m_rayFromTrans;
	btTransform	m_rayToTrans;
	btVector3	m_hitNormal;

	const btCollisionWorld*	m_world;
	btCollisionWorld::RayResultCallback&	m_resultCallback;

	btSingleRayCallback(const btVector3& rayFromWorld,const btVector3& rayToWorld,const btCollisionWorld* world,btCollisionWorld::RayResultCallback& resultCallback)
	:m_rayFromWorld(rayFromWorld),
	m_rayToWorld(rayToWorld),
	m_world(world),
	m_resultCallback(resultCallback)
	{
		m_rayFromTrans.setIdentity();
		m_rayFromTrans.setOrigin(m_rayFromWorld);
		m_rayToTrans.setIdentity();
		m_rayToTrans.setOrigin(m_rayToWorld);

		btVector3 rayDir = (rayToWorld-rayFromWorld);

		rayDir.normalize ();
		///what about division by zero? --> just set rayDirection[i] to INF/1e30
		m_rayDirectionInverse[0] = rayDir[0] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[0];
		m_rayDirectionInverse[1] = rayDir[1] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[1];
		m_rayDirectionInverse[2] = rayDir[2] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[2];
		m_signs[0] = m_rayDirectionInverse[0] < 0.0;
		m_signs[1] = m_rayDirectionInverse[1] < 0.0;
		m_signs[2] = m_rayDirectionInverse[2] < 0.0;

		m_lambda_max = rayDir.dot(m_rayToWorld-m_rayFromWorld);

	}
		
	virtual bool	process(const btBroadphaseProxy* proxy)
	{
		///terminate further ray tests, once the closestHitFraction reached zero
		if (m_resultCallback.m_closestHitFraction == btScalar(0.f))
			return false;

		btCollisionObject*	collisionObject = (btCollisionObject*)proxy->m_clientObject;

		//only perform raycast if filterMask matches
		if(m_resultCallback.needsCollision(collisionObject->getBroadphaseHandle())) 
		{
			//RigidcollisionObject* collisionObject = ctrl->GetRigidcollisionObject();
			//btVector3 collisionObjectAabbMin,collisionObjectAabbMax;
#if 0
#ifdef RECALCULATE_AABB
			btVector3 collisionObjectAabbMin,collisionObjectAabbMax;
			collisionObject->getCollisionShape()->getAabb(collisionObject->getWorldTransform(),collisionObjectAabbMin,collisionObjectAabbMax);
#else
			//getBroadphase()->getAabb(collisionObject->getBroadphaseHandle(),collisionObjectAabbMin,collisionObjectAabbMax);
			const btVector3& collisionObjectAabbMin = collisionObject->getBroadphaseHandle()->m_aabbMin;
			const btVector3& collisionObjectAabbMax = collisionObject->getBroadphaseHandle()->m_aabbMax;
#endif
#endif
			//btScalar hitLambda = m_resultCallback.m_closestHitFraction;
			//culling already done by broadphase
			//if (btRayAabb(m_rayFromWorld,m_rayToWorld,collisionObjectAabbMin,collisionObjectAabbMax,hitLambda,m_hitNormal))
			{
				m_world->rayTestSingle(m_rayFromTrans,m_rayToTrans,
					collisionObject,
						collisionObject->getCollisionShape(),
						collisionObject->getWorldTransform(),
						m_resultCallback);
			}
		}
		return true;
	}
};  
void Raytracer::updateRaytraceTexture(int textureIndex)	{
	unsigned char* pData = &textureBuffer[0];
	const btScalar w = (btScalar) textureSize[textureIndex];
	const btScalar h = (btScalar) textureSize[textureIndex];
	
	if (!pData || m_glutScreenWidth==0 || m_glutScreenHeight==0 || w==0 || h==0) return;
	if (m_cameraDistance<m_minCameraDistance)  {m_cameraDistance=m_minCameraDistance;updateCamera();}

	const btScalar screenRatio = (btScalar)m_glutScreenWidth/(btScalar)m_glutScreenHeight;

	const btVector3& SRC = m_cameraPosition;
	const btVector3& DST = m_cameraTargetPosition;
	
	//--From-------- void DemoApplication::updateCamera()---------------------
	btScalar rele = m_ele * btScalar(0.01745329251994329547);// rads per deg
	btScalar razi = m_azi * btScalar(0.01745329251994329547);// rads per deg
	btQuaternion rot = btQuaternion(m_cameraUp,razi);
	btVector3 eyePos(0,0,0);
	eyePos[m_forwardAxis] = -m_cameraDistance;
	btVector3 forward(eyePos[0],eyePos[1],eyePos[2]);
	if (forward.length2() < SIMD_EPSILON) forward.setValue(1.f,0.f,0.f);
	btVector3 right = m_cameraUp.cross(forward);
	btQuaternion roll(right,-rele);
	//-------------------------------------------------------------------------
	
	btTransform T; 
	T.setOrigin(SRC);
	T.setBasis(btMatrix3x3(rot) * btMatrix3x3(roll));	
	btVector3 rayDir = (DST-SRC);
	btScalar rayLength = rayDir.length()*2;
	rayLength = rayLength < 50 ? 50 : rayLength;
	rayDir.normalize();
	btVector3 lightNormal = rayDir;
	
				
	const btScalar tanFOVyHalf = tanf(btScalar(0.5)*m_orthoCameraFov); 
	const btScalar NMY = m_orthoCameraFocusDistance * tanFOVyHalf;
	const btScalar NMX = NMY * screenRatio;
	
	// Steps:
	const btScalar DMX = NMX/w;
	const btScalar DMY = NMY/h;
	
	btTransform TMX,TMY;	

	{
	btCollisionWorld::ClosestRayResultCallback cb(T.getOrigin(), T.getOrigin());
	cb.m_flags=0;
			
	btSingleRayCallback rayCB(T.getOrigin(),T.getOrigin(),m_collisionWorld,cb);

	unsigned char C;float Q;
	
	const float ambient = 0.25;
	const unsigned char ambientu = (unsigned char)(ambient*255.0);
	btScalar tempX,tempY;
	
	for (int y=0;y<h;y++)	{
			TMY = T;
			tempY = (h*0.5 - btScalar(y));
			TMY.getOrigin() = btVector3(TMY.getOrigin() + TMY.getBasis().getColumn(1) * tempY*DMY);
			
		for (int x=0;x<w;x++)	{
			TMX = TMY;
			tempX = (w*0.5 - btScalar(x));
			TMX.getOrigin() = btVector3(TMX.getOrigin() + TMX.getBasis().getColumn(0) * tempX*DMX);			
			
			cb.m_rayFromWorld = rayCB.m_rayFromWorld = TMX.getOrigin();	
			rayCB.m_rayFromTrans.setIdentity();					
			rayCB.m_rayFromTrans.setOrigin(TMX.getOrigin());

			TMX.getOrigin() = btVector3(TMX.getOrigin() + TMX.getBasis().getColumn(2) * rayLength);
	
			cb.m_rayToWorld = rayCB.m_rayToWorld = TMX.getOrigin();	
			rayCB.m_rayToTrans.setIdentity();				
			rayCB.m_rayToTrans.setOrigin(TMX.getOrigin());	
				
			
			 //region Fill 'cb' based on 'cb' and 'rayCB' (Bullet code unwrapped): does not detect soft bodies
			cb.m_closestHitFraction = btScalar(1.);
			cb.m_collisionObject=0;		
			
			rayCB.m_rayDirectionInverse[0] = rayDir[0] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[0];
			rayCB.m_rayDirectionInverse[1] = rayDir[1] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[1];
			rayCB.m_rayDirectionInverse[2] = rayDir[2] == btScalar(0.0) ? btScalar(1e30) : btScalar(1.0) / rayDir[2];
			rayCB.m_signs[0] = rayCB.m_rayDirectionInverse[0] < 0.0;
			rayCB.m_signs[1] = rayCB.m_rayDirectionInverse[1] < 0.0;
			rayCB.m_signs[2] = rayCB.m_rayDirectionInverse[2] < 0.0;
			rayCB.m_lambda_max = rayDir.dot(rayCB.m_rayToWorld-rayCB.m_rayFromWorld);

			/// use the broadphase to accelerate the search for objects, based on their aabb
			/// and for each object with ray-aabb overlap, perform an exact ray test
			#ifndef USE_BRUTEFORCE_RAYBROADPHASE
				m_collisionWorld->getBroadphase()->rayTest(rayCB.m_rayFromWorld,rayCB.m_rayToWorld,rayCB);
			#else
			for (int i = 0; i < this->getNumCollisionObjects() ; i++)
			{
				rayCB.process(m_collisionWorld->getCollisionObjectArray()[i]->getBroadphaseHandle());
			}	
			#endif //USE_BRUTEFORCE_RAYBROADPHASE
			
			//endregion
			
			if (cb.hasHit())	{
				Q = -(cb.m_hitNormalWorld.dot(lightNormal));
				if (Q<0) Q=0;
				Q+=ambient;
				if (Q>1) Q=1;
				C = (unsigned char) (Q * 255.0); 
				*pData++ = C;
			}
			else {
				*pData++ = ambientu;			
			}													
		}
	}
	}

	updateTexture(textureIndex);
}

void Raytracer::displayCallback() {
// Important: now this method is not "heart-beating": it's just called when glutPostRedisplay() is called.
// P.S. I don't know if stepSimulation() is called every frame: to maximize speed we could just
// call it once after initPhysics() as traytracing probably doesn't need it... (we have a collision
// world (not dynamic), so objects can't move by themselves) 

	if (m_curFrameTimer!=0)	{
		if (--m_curFrameTimer<=0) {
			m_curFrameTimer = 0;
			updateRaytraceTexture(1);	// casts rays and updates the texture
			//printf("updateRaytraceTexture(1); m_curFrameTimer=0\n");
		}
		else {
			if (m_curFrameTimer==framesInterval-1) {
				updateRaytraceTexture(0);	// casts rays and updates the texture
				//printf("updateRaytraceTexture(0); m_curFrameTimer: %1.0i\n",m_curFrameTimer);
			}	
			//else we're simply waiting around m_curFrameTimer, but we must display the opengl page as well
		}	
	}
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 
	

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();									
	glTranslatef(0.0f,0.0f,-1.1f);						
	
	glColor4f (1,1,1,1); 
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-1,1);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(1,1);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(1,-1);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(-1,-1);
	glEnd();
	

	glFlush();
	glutSwapBuffers();
	
	if (m_curFrameTimer>0) glutPostRedisplay();
}
void Raytracer::specialKeyboard(int key, int x, int y)	{
	(void)x;
	(void)y;

	switch (key) 
	{
	case GLUT_KEY_F1:
		{

			break;
		}

	case GLUT_KEY_F2:
		{

			break;
		}


	case GLUT_KEY_LEFT : {
			m_curFrameTimer = framesInterval;
			stepLeft(); 
		}
		break;
	case GLUT_KEY_RIGHT : {
			m_curFrameTimer = framesInterval;
			stepRight(); 
		}
		break;
	case GLUT_KEY_UP : {
			m_curFrameTimer = framesInterval;
			stepFront(); 
		}
		break;
	case GLUT_KEY_DOWN : {
			m_curFrameTimer = framesInterval;
			stepBack(); 
		}
		break;
	case GLUT_KEY_PAGE_UP :  {
		m_curFrameTimer = framesInterval;
		//zoomIn(); 
		m_orthoCameraFocusDistance-=0.25f;
		if (m_orthoCameraFocusDistance<1.0) m_orthoCameraFocusDistance=1.0;
		}
		break;
	case GLUT_KEY_PAGE_DOWN :  {
		m_curFrameTimer = framesInterval;	
		//zoomOut(); 
		m_orthoCameraFocusDistance+=0.25f;
		if (m_orthoCameraFocusDistance>100.0) m_orthoCameraFocusDistance=100.0;				
		}
		break;
	case GLUT_KEY_HOME : toggleIdle(); break;
	default:
		//        std::cout << "unused (special) key : " << key << std::endl;
		break;
	}

	glutPostRedisplay();

}

//================================================================
// class btCollisionShapeGarbageCollector
//================================================================
void btCollisionShapeGarbageCollector::cleanList()	{
	/*
	1) removeDoubles
	2) find subshapes (recursive)
	3) add subshapes
	2+3) flatten shape hierarchy
	4) remove doubles	
	*/
	removeDoublesAndNulls();
	flattenHierarchy(true);
}
void btCollisionShapeGarbageCollector::removeDoublesAndNulls()	{
		for (int i=0;i<size()-1;i++)	{
			btCollisionShape* item = operator[](i);
			if (!item) {
				Base::swap(i,size()-1);
				Base::pop_back();
				--i;
				continue;
			}
			for (int j=i+1;j<size();j++)	{
				if (item == operator[](j)) {
					Base::swap(j,size()-1);
					Base::pop_back();
					--j;
				}
			}
		}
}
void btCollisionShapeGarbageCollector::flattenHierarchy(bool dontInsertDoublesAndNulls)	{
	const int parentShapeArraySize = size();
	for (int i = 0; i < parentShapeArraySize;i++)	{
		addChildShapesOf(operator[](i),dontInsertDoublesAndNulls,false);
	}
}
void btCollisionShapeGarbageCollector::addChildShapesOf(btCollisionShape* shape,bool dontInsertDoublesAndNulls,bool addThePassedShapeIfLeafShape)	{
		if (!shape) {
			if (addThePassedShapeIfLeafShape) push_back(shape,dontInsertDoublesAndNulls);
			return;
		}	
		switch (shape->getShapeType())	{
			case COMPOUND_SHAPE_PROXYTYPE:
				{
					btCompoundShape* cs = static_cast <btCompoundShape*> (shape);
					for (int j=0,childSz=cs->getNumChildShapes();j<childSz;j++)	{
						addChildShapesOf(cs->getChildShape(j),dontInsertDoublesAndNulls,true);
					}
				}
			break;
			case UNIFORM_SCALING_SHAPE_PROXYTYPE:
				{
					btUniformScalingShape* s = static_cast <btUniformScalingShape*> (shape);
					addChildShapesOf(s->getChildShape(),dontInsertDoublesAndNulls,true);
				}
			break;
			case SCALED_TRIANGLE_MESH_SHAPE_PROXYTYPE:
				{
					btScaledBvhTriangleMeshShape* s = static_cast <btScaledBvhTriangleMeshShape*> (shape);
					addChildShapesOf(s->getChildShape(),dontInsertDoublesAndNulls,true);
				}
			break;
			default:
				if (addThePassedShapeIfLeafShape) push_back(shape,dontInsertDoublesAndNulls);
			break;
		}
}
// Debuggable:
void btCollisionShapeGarbageCollector::deleteAll()	{		
		cleanList();
		//#define DEBUGME
		#ifdef DEBUGME
		printf("Deleting %1.0i\n",size());
		#endif //DEBUGME
		for (int i=size()-1;i>=0;i--)	{
			btCollisionShape* s = operator[](i);
			if (s) {
				#ifdef DEBUGME
				printf("%1.0i) deleting: ",size()-1-i);
				printf(s->getName());
				printf("\n");
				#endif //DEBUGME
				delete s;s=NULL;
			}
		}
		Base::clear();
		#ifdef DEBUGME
		#undef DEBUGME
		#endif
	}
//=================================================================

 //region SallyIDE Text Editor Settings
//Caret=(0,75)
//Bookmarks=(471)
//endregion
